//----------------------------------------------------------------
//
//-- Filename: crypto.c
//
//-- $Date: 2008/03/13 15:10:27 $
//-- $Revision: 1.4 $
//-- $Name:  $
//
//----------------------------------------------------------------

#include "ils/internal/internal.h"
#include "ils/util/crypto.h"
#include "ils/util/hexdump.h"

#if defined(HAVE_OPENSSL_CRYPTO_H)
#  include <openssl/crypto.h>
#endif

#if defined(HAVE_OPENSSL_AES_H)
#  include <openssl/aes.h>
#else
#  include "ils/internal/aes.h"
#endif

#define CIPHER_ALG_MASK		0x000000FF
#define CIPHER_ALG_AES		0x00000001

#define CIPHER_MODE_MASK	0x0000FF00
#define CIPHER_MODE_ECB		0x00000100
#define CIPHER_MODE_CBC		0x00000200
#define CIPHER_MODE_CFB		0x00000400

#define CIPHER_PAD_MASK		0x00FF0000
#define CIPHER_PAD_NONE		0x00010000
#define CIPHER_PAD_PKCS5	0x00020000

#define CIPHER_AES_ECB_NONE			CIPHER_ALG_AES|CIPHER_MODE_ECB|CIPHER_PAD_NONE
#define CIPHER_AES_ECB_PKCS5		CIPHER_ALG_AES|CIPHER_MODE_ECB|CIPHER_PAD_PKCS5

#define CIPHER_AES_CBC_NONE			CIPHER_ALG_AES|CIPHER_MODE_CBC|CIPHER_PAD_NONE
#define CIPHER_AES_CBC_PKCS5		CIPHER_ALG_AES|CIPHER_MODE_CBC|CIPHER_PAD_PKCS5

#define CIPHER_AES_CFB_NONE			CIPHER_ALG_AES|CIPHER_MODE_CFB|CIPHER_PAD_NONE

#define CIPHER_ENCRYPT	1
#define CIPHER_DECRYPT	2

typedef struct {
	int cipher;
	char def[32];
	unsigned char key[32];
	unsigned char iv[16];
	int ctr;
	int mode;
	union {
		AES_KEY	aesCTX;
	} CTX;
} _CRYPTO_CIPHER;

int _crypto_create(void **cipher, const char *definition, const unsigned char *key, const unsigned char *iv, int cmode) {
	_CRYPTO_CIPHER *_cipher;
	char *alg=NULL,*mode=NULL,*pad=NULL;
	int keylen=0;
	#if defined(HAS_STRTOK_R)
	char *reent;
	#endif
	
	if(cipher==NULL) return ERR_BADHANDLE;
	*cipher=NULL;
	if(definition==NULL || key==NULL || iv==NULL) return ERR_PARMS;
	
	if( (_cipher=ilsMalloc(sizeof(_CRYPTO_CIPHER)))==NULL ) {
		return ERR_MALLOC;
	}
	ilsMemset(_cipher,0,sizeof(_CRYPTO_CIPHER));
	
	strncpy(_cipher->def,definition,31);
	ilsMemcpy(_cipher->iv,iv,AES_BLOCK_SIZE);
	_cipher->mode=cmode;
	
	alg=strtok_r(_cipher->def,"/-",&reent);
	mode=strtok_r(NULL,"/-",&reent);
	pad=strtok_r(NULL,"/-",&reent);
	
	if(strncasecmp(alg,"aes",3)==0) { // AES
		keylen=128;
		if(strlen(alg)>3) {
			keylen=atoi(alg+3);
			if( (keylen!=128) && (keylen!=192) && (keylen!=256) ) {
				ilsFree(_cipher);
				return ERR_CRYPTO_BAD_KEYLEN;
			}
		}
		ilsMemcpy(_cipher->key,key,keylen/8);
		_cipher->cipher|=CIPHER_ALG_AES;
	} else {
		ilsFree(_cipher);
		return ERR_CRYPTO_BAD_ALGORITHM;
	}
	
	if(mode==NULL || strcasecmp(mode,"ecb")==0) {
		_cipher->cipher|=CIPHER_MODE_ECB;
	} else if(mode==NULL || strcasecmp(mode,"cbc")==0) {
		_cipher->cipher|=CIPHER_MODE_CBC;
	} else if(mode==NULL || strcasecmp(mode,"cfb")==0) {
		_cipher->cipher|=CIPHER_MODE_CFB;
	} else {
		ilsFree(_cipher);
		return ERR_CRYPTO_BAD_MODE;
	}
	
	if(pad==NULL || strcasecmp(pad,"none")==0) {
		_cipher->cipher|=CIPHER_PAD_NONE;
	} else if(strcasecmp(pad,"pkcs5")==0) {
		if(_cipher->cipher&CIPHER_MODE_CFB) {
			ilsFree(_cipher);
			return ERR_CRYPTO_BAD_PADDING;		
		} else {
			_cipher->cipher|=CIPHER_PAD_PKCS5;
		}
	} else {
		ilsFree(_cipher);
		return ERR_CRYPTO_BAD_PADDING;
	}
	
	if( (_cipher->cipher&CIPHER_ALG_AES)>0 ) {
		if(cmode==CIPHER_ENCRYPT || (_cipher->cipher&CIPHER_MODE_CFB)>0 ) {
			if( AES_set_encrypt_key( (const unsigned char *)key, keylen, &_cipher->CTX.aesCTX) != 0 ) {
				return ERR_CRYPTO_UNKNOWN;
			}
		} else {
			if( AES_set_decrypt_key( (const unsigned char *)key, keylen, &_cipher->CTX.aesCTX) != 0 ) {
				return ERR_CRYPTO_UNKNOWN;
			}
		}
	}
	
	strncpy(_cipher->def,definition,31);

	*cipher=_cipher;
	return 0;
}

int encryptor_create(void **cipher, const char *definition, const unsigned char *key, const unsigned char *iv) {
	return _crypto_create(cipher,definition,key,iv,CIPHER_ENCRYPT);
}

int decryptor_create(void **cipher, const char *definition, const unsigned char *key, const unsigned char *iv) {
	return _crypto_create(cipher,definition,key,iv,CIPHER_DECRYPT);
}

int encryptor_delete(void *cipher) {
	if(cipher==NULL) return ERR_BADHANDLE;
	ilsFree(cipher);
	return 0;
}

int decryptor_delete(void *cipher) {
	if(cipher==NULL) return ERR_BADHANDLE;
	ilsFree(cipher);
	return 0;
}

int encryptor_encrypt(void *cipher, const void *in, int inlen, void *out, int *outlen ) {
	return encryptor_encrypt_ex(cipher,in,inlen,out,outlen,0);
}

int encryptor_encrypt_ex(void *cipher, const void *in, int inlen, void *out, int *outlen, int options ) {

	_CRYPTO_CIPHER *_cipher=cipher;
	int rem=0;
	int bytes_leftover=0;
	int bytes_to_encrypt=0;
	
	if(cipher==NULL) return ERR_BADHANDLE;
	if(_cipher->mode!=CIPHER_ENCRYPT) return ERR_CRYPTO_WRONG_MODE;
		
	switch( _cipher->cipher ) {
		case CIPHER_AES_ECB_NONE:
			if( inlen%AES_BLOCK_SIZE != 0 )
				return ERR_CRYPTO_BAD_BLOCKSIZE;
			if(inlen>*outlen) 
				return ERR_CRYPTO_INSUF_BUFFER;
			while( rem != inlen ) {
				AES_ecb_encrypt( 	(unsigned char *)in+rem, 
									(unsigned char *)out+rem, 
									&_cipher->CTX.aesCTX, 
									AES_ENCRYPT );
				rem += AES_BLOCK_SIZE;
			}
			*outlen=inlen;
			break;
		case CIPHER_AES_ECB_PKCS5:
			bytes_leftover=inlen%AES_BLOCK_SIZE;
			if( (options&CRYPTO_OPT_NOPAD) ) {
				if( inlen%AES_BLOCK_SIZE != 0 )
					return ERR_CRYPTO_BAD_BLOCKSIZE;
				if(inlen>*outlen) 
					return ERR_CRYPTO_INSUF_BUFFER;
			} else {
				if((inlen+AES_BLOCK_SIZE-bytes_leftover)>*outlen) 
					return ERR_CRYPTO_INSUF_BUFFER;
			}
			
			bytes_to_encrypt=inlen-bytes_leftover;
			
			while( rem != bytes_to_encrypt ) {
				AES_ecb_encrypt( 	(unsigned char *)in+rem, 
									(unsigned char *)out+rem, 
									&_cipher->CTX.aesCTX, 
									AES_ENCRYPT );
				rem += AES_BLOCK_SIZE;
			}
			if( (options&CRYPTO_OPT_NOPAD)==0) {
				ilsMemset(((unsigned char *)out)+bytes_to_encrypt,bytes_leftover,AES_BLOCK_SIZE);
				if(bytes_leftover>0)
					ilsMemcpy(((unsigned char *)out)+bytes_to_encrypt,((unsigned char *)in)+bytes_to_encrypt,bytes_leftover);
				AES_ecb_encrypt( 	(unsigned char *)in+bytes_to_encrypt, 
									(unsigned char *)out+bytes_to_encrypt, 
									&_cipher->CTX.aesCTX, 
									AES_ENCRYPT );
				*outlen=inlen+AES_BLOCK_SIZE-bytes_leftover;
			} else {
				*outlen=inlen;
			}
			break;
		case CIPHER_AES_CBC_NONE:
			if( inlen%AES_BLOCK_SIZE != 0 )
				return ERR_CRYPTO_BAD_BLOCKSIZE;
			if(inlen>*outlen) 
				return ERR_CRYPTO_INSUF_BUFFER;

			AES_cbc_encrypt( 	(unsigned char *)in, 
								(unsigned char *)out, 
								inlen, 
								&_cipher->CTX.aesCTX, 
								_cipher->iv, 
								AES_ENCRYPT );
			*outlen=inlen;
			break;
		case CIPHER_AES_CBC_PKCS5:
			bytes_leftover=inlen%AES_BLOCK_SIZE;
			
			if( (options&CRYPTO_OPT_NOPAD) ) {
				if( inlen%AES_BLOCK_SIZE != 0 )
					return ERR_CRYPTO_BAD_BLOCKSIZE;
				if(inlen>*outlen) 
					return ERR_CRYPTO_INSUF_BUFFER;
			} else {
				if((inlen+AES_BLOCK_SIZE-bytes_leftover)>*outlen) 
					return ERR_CRYPTO_INSUF_BUFFER;
			}
			
			bytes_to_encrypt=inlen-bytes_leftover;

			AES_cbc_encrypt( 	(unsigned char *)in, 
								(unsigned char *)out, 
								bytes_to_encrypt,
								&_cipher->CTX.aesCTX,
								_cipher->iv,
								AES_ENCRYPT );
			
			if( (options&CRYPTO_OPT_NOPAD)==0) {
				ilsMemset(((unsigned char *)out)+bytes_to_encrypt,AES_BLOCK_SIZE-bytes_leftover,AES_BLOCK_SIZE);
				if(bytes_leftover>0)
					ilsMemcpy(((unsigned char *)out)+bytes_to_encrypt,((unsigned char *)in)+bytes_to_encrypt,bytes_leftover);
				AES_cbc_encrypt( 	(unsigned char *)out+bytes_to_encrypt, 
									(unsigned char *)out+bytes_to_encrypt, 
									AES_BLOCK_SIZE,
									&_cipher->CTX.aesCTX, 
									_cipher->iv,
									AES_ENCRYPT );
				*outlen=inlen+AES_BLOCK_SIZE-bytes_leftover;
			} else {
				*outlen=inlen;
			}
			break;
		case CIPHER_AES_CFB_NONE:
			if(inlen>*outlen) 
				return ERR_CRYPTO_INSUF_BUFFER;
			AES_cfb128_encrypt( (unsigned char *)in, 
								(unsigned char *)out, 
								inlen, 
								&_cipher->CTX.aesCTX, 
								_cipher->iv, 
								&_cipher->ctr,
								AES_ENCRYPT );
			*outlen=inlen;
			break;
	}
	
	return 0;
}

int decryptor_decrypt(void *cipher, const void *in, int inlen, void *out, int *outlen ) {
	return decryptor_decrypt_ex(cipher,in,inlen,out,outlen,0);
}

int decryptor_decrypt_ex(void *cipher, const void *in, int inlen, void *out, int *outlen, int options ) {
	_CRYPTO_CIPHER *_cipher=cipher;
	int rem=0;
	
	if(cipher==NULL) return ERR_BADHANDLE;
	if(_cipher->mode!=CIPHER_DECRYPT) return ERR_CRYPTO_WRONG_MODE;
		
	switch( _cipher->cipher ) {
		case CIPHER_AES_ECB_NONE:
			if( inlen%AES_BLOCK_SIZE != 0 )
				return ERR_CRYPTO_BAD_BLOCKSIZE;
			if(inlen>*outlen) 
				return ERR_CRYPTO_INSUF_BUFFER;
			while( rem != inlen ) {
				AES_ecb_encrypt( 	(unsigned char *)in+rem, 
									(unsigned char *)out+rem, 
									&_cipher->CTX.aesCTX, 
									AES_DECRYPT );
				rem += AES_BLOCK_SIZE;
			}
			*outlen=inlen;
			break;
		case CIPHER_AES_ECB_PKCS5:
			if( inlen%AES_BLOCK_SIZE != 0 )
				return ERR_CRYPTO_BAD_BLOCKSIZE;
			if(inlen>*outlen) 
				return ERR_CRYPTO_INSUF_BUFFER;
			
			while( rem != inlen ) {
				AES_ecb_encrypt( 	(unsigned char *)in+rem, 
									(unsigned char *)out+rem, 
									&_cipher->CTX.aesCTX, 
									AES_DECRYPT );
				rem += AES_BLOCK_SIZE;
			}
			if( (options&CRYPTO_OPT_NOPAD)==0) {
				*outlen=inlen-((unsigned char *)out)[inlen-1];
				((unsigned char *)out)[*outlen]=0;
			} else {
				*outlen=inlen;
			}
			break;
		case CIPHER_AES_CBC_NONE:
			if( inlen%AES_BLOCK_SIZE != 0 )
				return ERR_CRYPTO_BAD_BLOCKSIZE;
			if(inlen>*outlen) 
				return ERR_CRYPTO_INSUF_BUFFER;

			AES_cbc_encrypt( 	(unsigned char *)in, 
								(unsigned char *)out, 
								inlen, 
								&_cipher->CTX.aesCTX, 
								_cipher->iv, 
								AES_DECRYPT );
			*outlen=inlen;
			break;
		case CIPHER_AES_CBC_PKCS5:
			if( inlen%AES_BLOCK_SIZE != 0 )
					return ERR_CRYPTO_BAD_BLOCKSIZE;
				if(inlen>*outlen) 
					return ERR_CRYPTO_INSUF_BUFFER;

				AES_cbc_encrypt( 	(unsigned char *)in, 
									(unsigned char *)out, 
									inlen,
									&_cipher->CTX.aesCTX, 
									_cipher->iv, 
									AES_DECRYPT ); 
				if( (options&CRYPTO_OPT_NOPAD)==0) {
					if(((unsigned char *)out)[inlen-1]<1 || ((unsigned char *)out)[inlen-1]>16) {
						return ERR_CRYPTO_BAD_PADDING;
					}
					*outlen=inlen-((unsigned char *)out)[inlen-1];
					((unsigned char *)out)[*outlen]=0;
				} else {
					*outlen=inlen;
				}
			break;
		case CIPHER_AES_CFB_NONE:
			if(inlen>*outlen) 
				return ERR_CRYPTO_INSUF_BUFFER;
			AES_cfb128_encrypt( (unsigned char *)in, 
								(unsigned char *)out, 
								inlen, 
								&_cipher->CTX.aesCTX, 
								_cipher->iv, 
								&_cipher->ctr,
								AES_DECRYPT );
			*outlen=inlen;
			break;
	}
	
	return 0;
}

int encryptor_encrypt_file(void *cipher, const char *src, const char *dst ) {
	FILE *rp, *wp;
	int inlen, outlen, ret;
	char in[512], out[512+32];
	
	if( (rp=fopen(src,"rb"))==NULL ) {
		return ERR_FOPEN_FAILED;
	}
	if( (wp=fopen(dst,"wb"))==NULL ) {
		fclose(rp);
		return ERR_FOPEN_FAILED;
	}

	while( (inlen = fread(in, 1, 512, rp))==512 ) {
		outlen=512+32;
		if( (ret=encryptor_encrypt_ex(cipher,in,inlen,out,&outlen,CRYPTO_OPT_NOPAD)) !=0 ) {
			fclose(rp);
			fclose(wp);
			return ret;
		}
		fwrite(out, 1, outlen, wp);
	}
	outlen=512+32;
	if( (ret=encryptor_encrypt_ex(cipher,in,inlen,out,&outlen,0)) !=0 ) {
		fclose(rp);
		fclose(wp);
		return ret;
	}
	fwrite(out, 1, outlen, wp);
   
	fclose(rp);
	fclose(wp);	
	return 0;
}

int decryptor_decrypt_file(void *cipher, const char *src, const char *dst ) {
	FILE *rp, *wp;
	int inlen, outlen, ret;
	char in[512], out[512+32];
	
	if( (rp=fopen(src,"rb"))==NULL ) {
		return ERR_FOPEN_FAILED;
	}
	if( (wp=fopen(dst,"wb"))==NULL ) {
		fclose(rp);
		return ERR_FOPEN_FAILED;
	}

	while( (inlen = fread(in, 1, 512, rp))==512 ) {
		outlen=512+32;
		if( (ret=decryptor_decrypt_ex(cipher,in,inlen,out,&outlen,CRYPTO_OPT_NOPAD)) !=0 ) {
			fclose(rp);
			fclose(wp);
			return ret;
		}
		fwrite(out, 1, outlen, wp);
	}
	outlen=512+32;
	if(inlen>0) {
		if( (ret=decryptor_decrypt_ex(cipher,in,inlen,out,&outlen,0)) !=0 ) {
			fclose(rp);
			fclose(wp);
			return ret;
		}
		fwrite(out, 1, outlen, wp);
	}
   
	fclose(rp);
	fclose(wp);	
	return 0;
}
